import xml.etree.ElementTree as et

import attr
from . import encoder as default_encoder
from . import exceptions as exc
from . import mappers


def model(maybe_cls=None, name=None, ns=None, ns_map=None, order=None, **kwargs):
    """
    The decorator maps a class to an XML element. It uses a class
    name as a default element name. The default name can be altered using the decorator
    `name` argument. The `ns` argument defines a namespace of the element. By default
    class fields are serialized in in-class definition order. The order can be altered
    using `order` attribute.

    Internally the decorator adds a `dunder <https://wiki.python.org/moin/DunderAlias>`_\
    attribute.

    :param maybe_cls: decorated class if it is used as `@model` or `None` if it is used as `@model()`
    :param str name: model name. If `None` class name will be used
    :param str ns: element namespace. If `None` empty namespace will be used or if the model
               is a nested one - namespace is inherited from the containing model
    :param dict ns_map: mapping from a namespace prefix to a full name. It is applied to the current model
                        and it's elements and all nested models
    :param tuple order: class fields serialization order. If `None` in-class definition order is used
    :param kwargs: arguments that will be passed to :py:func:`attr.ib`
    """

    def decorator(cls):
        cls = attr.attrs(cls, **kwargs)
        if order:
            for element_name in order:
                if not hasattr(getattr(cls, '__attrs_attrs__'), element_name):
                    raise AssertionError("order element '{}' not declared in model".format(element_name))

        cls.__paxb_attrs__ = (name, ns, ns_map, order)

        return cls

    # maybe_cls's type depends on the usage of the decorator.  It's a class
    # if it's used as `@model` but ``None`` if used as `@model()`.
    if maybe_cls is None:
        return decorator
    else:
        return decorator(maybe_cls)


def attribute(name=None, **kwargs):
    """
    The Function maps a class field to an XML attribute. The field name is used
    as a default attribute name. The default name can be altered using the `name` argument.

    :param str name: attribute name. If `None` field name will be used
    :param kwargs: arguments that will be passed to :py:func:`attr.ib`
    """

    has_default = 'default' in kwargs or 'factory' in kwargs
    required = not has_default

    attrib = attr.attrib(**kwargs)
    attrib.metadata['paxb.mapper'] = mappers.AttributeXmlMapper(name, required)

    return attrib


def field(name=None, ns=None, ns_map=None, idx=None, **kwargs):
    """
    The Function maps a class field to an XML element. The field name is used
    as a default element name. The default name can be altered using `name` argument.
    The `ns` argument defines the namespace of the element.

    Internally the decorator adds some metainformation to :py:attr:`attr.ib.metadata`.

    :param str name: element name. If `None` field name will be used
    :param str ns: element namespace. If `None` the namespace is inherited from the containing model
    :param dict ns_map: mapping from a namespace prefix to a full name.
    :param int idx: element index in the xml document. If `None` 1 is used
    :param kwargs: arguments that will be passed to :py:func:`attr.ib`
    """

    has_default = 'default' in kwargs or 'factory' in kwargs
    required = not has_default

    attrib = attr.attrib(**kwargs)
    attrib.metadata['paxb.mapper'] = mappers.FieldXmlMapper(name, ns, ns_map, idx, required)

    return attrib


def nested(cls, name=None, ns=None, ns_map=None, idx=None, **kwargs):
    """
    The Function maps a class to an XML element. `nested` is used when a :py:func:`paxb.model`
    decorated class contains another one as a field.

    :param cls: nested object class. `cls` must be an instance of :py:func:`paxb.model` decorated class
    :param str name: element name. If `None` model decorator `name` attribute will be used
    :param str ns: element namespace. If `None` model decorator ns attribute will be used
    :param dict ns_map: mapping from a namespace prefix to a full name. It is applied to the current model
                        and it's elements and all nested models
    :param int idx: element index in the xml document. If `None` 1 is used
    :param kwargs: arguments that will be passed to :py:func:`attr.ib`
    """

    if not isinstance(cls, type):
        raise TypeError("Passed object must be a class")

    required = 'default' not in kwargs

    attrib = attr.attrib(**kwargs)
    attrib.metadata['paxb.mapper'] = mappers.ModelXmlMapper(cls, name, ns, ns_map, idx, required)

    return attrib


def wrapper(path, wrapped, ns=None, ns_map=None, idx=None):
    """
    The Function is used to map a class field to an XML element that is contained by a subelement.

    :param str path: full path to the `wrapped` element. Element names are separated by slashes
    :param wrapped: a wrapped element
    :param str ns: element namespace. If `None` the namespace is inherited from the containing model
    :param dict ns_map: mapping from a namespace prefix to a full name. It is applied to the current model
                        and it's elements and all nested models
    :param int idx: element index in the xml document. If `None` 1 is used
    """

    wrapped.metadata['paxb.mapper'] = mappers.WrapperXmlMapper(path, wrapped.metadata['paxb.mapper'], ns, ns_map, idx)

    return wrapped


def as_list(wrapped):
    """
    The Function maps a class list field to an XML element list. Wrapped element
    can be field or nested model.

    :param wrapped: list element type
    """

    wrapped.metadata['paxb.mapper'] = mappers.ListXmlWrapper(wrapped.metadata['paxb.mapper'])

    return wrapped


def from_xml(cls, xml, envelope=None, name=None, ns=None, ns_map=None, required=True):
    """
    Deserializes xml string to object of `cls` type. `cls` must be a :py:func:`paxb.model` decorated class.

    :param cls: class the deserialized object is instance of
    :param xml: xml string or xml tree to deserialize the object from
    :type xml: :py:class:`str` or :py:class:`xml.etree.ElementTree.ElementTree`
    :param str envelope: root tag where the serializing object will be looked for
    :param str name: name of the serialized object element. If `None` model decorator `name` argument will be used
    :param str ns: namespace of the serialized object element. If `None` model decorator `ns` argument will be used
    :param dict ns_map: mapping from a namespace prefix to a full name
    :param bool required: is the serialized object element required. If element not found and `required` is ``True``
           :py:exc:`paxb.exceptions.DeserializationError` will be raised otherwise ``None`` is returned
    :return: deserialized object
    """

    if isinstance(xml, str):
        root = et.Element(None)
        root.append(et.fromstring(xml))
    else:
        root = xml

    if envelope:
        root = root.find(envelope, ns_map)
        if root is None:
            raise exc.DeserializationError("required element '{}' not found".format(envelope))
        full_path = tuple(envelope.split('/'))
    else:
        full_path = ()

    return mappers.ModelXmlMapper(cls, name, ns, ns_map, required=required).obj(root, full_path=full_path)


def to_xml(obj, envelope=None, name=None, ns=None, ns_map=None, encoder=default_encoder, **kwargs):
    """
    Serializes a ``paxb`` model object to an xml string. Object must be an instance
    of a :py:func:`paxb.model` decorated class.

    :param obj: object to be serialized
    :param str envelope: root tag name the serialized object element will be added inside.
                     If ``None`` object element will be a root
    :param str name: name of the serialized object element. If `None` model decorator `name` argument will be used
    :param str ns: namespace of the serialized object element. If `None` model decorator `ns` argument will be used
    :param dict ns_map: mapping from a namespace prefix to a full name.
    :param encoder: value encoder. If ``None`` :py:func:`paxb.encoder.encode` is used
    :param kwargs: arguments that will be passed to :py:func:`xml.etree.ElementTree.tostring` method
    :return: serialized object xml string
    :rtype: :py:class:`bytes` or :py:class:`str`
    """

    if ns_map:
        for alias, uri in ns_map.items():
            et.register_namespace(alias, uri)

    root = et.Element(envelope)
    obj = mappers.ModelXmlMapper(obj.__class__, name, ns, ns_map).xml(obj, root, encoder=encoder)

    if envelope is None:
        root = obj

    if root is not None:
        return et.tostring(root, **kwargs)
    else:
        return None



"""
``paxb`` is a library that provides an API for mapping between XML documents and Python objects.

``paxb`` library implements the following functionality:

    - Deserialize XML documents to Python objects
    - Validate deserialized data
    - Access and update Python object fields
    - Serialize Python objects to XML documents

paxb provides an efficient way of mapping between an XML document and a Python object. Using paxb
developers can write less boilerplate code emphasizing on application business logic.

Since paxb based on `attrs <https://www.attrs.org/en/stable/index.html>`_ library paxb and attrs
API can be mixed together.
"""

from .__about__ import (
    __title__,
    __description__,
    __url__,
    __version__,
    __author__,
    __email__,
    __license__
)
from .paxb import (
    as_list,
    attribute,
    field,
    from_xml,
    nested,
    model,
    to_xml,
    wrapper,
)
from . import exceptions as exc


# shortcuts
attr = attribute
wrap = wrapper
lst = as_list


__all__ = [
    '__title__',
    '__description__',
    '__url__',
    '__version__',
    '__author__',
    '__email__',
    '__license__',

    'as_list',
    'attr',
    'attribute',
    'exc',
    'field',
    'from_xml',
    'lst',
    'nested',
    'model',
    'to_xml',
    'wrap',
    'wrapper',
]



__title__ = 'paxb'
__description__ = 'Python Architecture for XML Binding'
__url__ = 'https://github.com/dapper91/paxb'

__version__ = '0.2.1'

__author__ = 'Dmitry Pershin'
__email__ = 'dapper91@mail.ru'

__license__ = 'Public Domain License'



"""
Package errors.
"""


class BaseError(Exception):
    """
    Base package exception. All package exception are inherited from it.
    """


class SerializationError(BaseError):
    """
    Serialization error. Raised when any serialization error occurs.
    """


class DeserializationError(BaseError):
    """
    Deserialization error. Raised when any deserialization error occurs.
    """



"""
The module implements mappers and mapping routine functions. Mapper is a class that implements logic
for mapping xml elements to python object fields and vise versa.
"""

import abc
import collections
import operator as op
import xml.etree.ElementTree as et

import attr

from . import exceptions as exc
from . import encoder as default_encoder


def get_attrs(cls):
    """
    Returns all paxb attributes of a class decorated with :py:func:`paxb.model` decorator.

    :param cls: :py:func:`paxb.model` decorated class
    :return: paxb class attributes
    """

    if not isinstance(cls, type):
        raise TypeError("Passed object must be a class")

    attrs = getattr(cls, '__paxb_attrs__', None)
    if attrs is None:
        raise TypeError("{cls!r} is not a paxb model-decorated class.".format(cls=cls))

    return attrs


def tag_name(ns, name, idx=None):
    tag = '{}:{}'.format(ns, name) if ns else name

    if idx is not None:
        tag += '[{}]'.format(idx)

    return tag


def qname(ns, name):
    return '{{{}}}{}'.format(ns, name) if ns else name


def drop_nones(d):
    """
    Drop none values from the dict
    """
    return {k: v for k, v in d.items() if v is not None}


def first(*args):
    """
    Returns first not `None` argument.
    """

    for item in args:
        if item is not None:
            return item


def merge_dicts(*dicts):
    """
    Merges dicts in reversed order.

    :param dicts: dicts to be merged
    :return: merged dict
    """

    result = {}
    for d in reversed(dicts):
        if d:
            result.update(d)

    return result


def reorder(fields, order, key):
    """
    Reorders `fields` list sorting its elements in order they appear in `order` list.
    Elements that are not defined in `order` list keep the original order.

    :param fields: elements to be reordered
    :param order: iterable that defines a new order
    :param key: a function of one argument that is used to extract a comparison key from each element in `fields`
    :return: reordered elements list
    """

    ordered = collections.OrderedDict()
    for field in fields:
        ordered[key(field)] = field

    for ord in reversed(order or ()):
        ordered.move_to_end(ord, last=False)

    return ordered.values()


class Mapper(abc.ABC):
    """
    Base mapper class. All mappers are inherited from it.
    """

    @abc.abstractmethod
    def xml(self, obj, root, name=None, ns=None, ns_map=None, idx=None, encoder=default_encoder):
        """
        Serialization method.

        :param obj: object to be serialized
        :param root: root element the object will be added inside
        :type root: :py:class:`xml.etree.ElementTree.Element`
        :param str name: element name
        :param str ns: element namespace
        :param dict ns_map: mapping from namespace prefix to full name
        :param int idx: element index in the xml tree
        :param encoder: value encoder
        :return: added xml tree node
        """

    @abc.abstractmethod
    def obj(self, xml, name=None, ns=None, ns_map=None, idx=None, full_path=()):
        """
        Deserialization method.

        :param xml: xml tree to deserialize the object from
        :type xml: :py:class:`xml.etree.ElementTree.Element`
        :param str name: element name
        :param str ns: element namespace
        :param dict ns_map: mapping from namespace prefix to full name
        :param int idx: element index in the xml tree
        :param tuple full_path: full path to the current element
        :return: deserialized object
        """


# TODO: implement attribute namespaces
class AttributeXmlMapper(Mapper):
    """
    Attribute to XMl mapper. Implements methods for mapping an xml attribute to a python object and vise versa.
    """

    def __init__(self, name=None, required=True):
        self.name = name
        self.required = required

    def xml(self, obj, root, name=None, ns=None, ns_map=None, _=None, encoder=default_encoder):
        name = first(self.name, name)

        if obj is None:
            if self.required:
                raise exc.SerializationError("required attribute '{}' is not set".format(name))
            else:
                return None

        attrib = encoder.encode(obj)
        root.attrib[name] = attrib

        return attrib

    def obj(self, xml, name=None, ns=None, ns_map=None, _=None, full_path=()):
        name = first(self.name, name)

        attribute = xml.get(name)
        if attribute is None and self.required:
            raise exc.DeserializationError("required attribute '/{}' not found".format('/'.join(full_path + (name, ))))

        return attribute


class FieldXmlMapper(Mapper):
    """
    XML element mapper. Implements methods for mapping an xml element text data to a python object and vise versa.
    """

    def __init__(self, name, ns=None, ns_map=None, idx=None, required=True):
        self.name = name
        self.ns = ns
        self.ns_map = ns_map
        self.idx = idx
        self.required = required

    def xml(self, obj, root, name=None, ns=None, ns_map=None, idx=None, encoder=default_encoder):
        name = first(self.name, name)
        ns = first(self.ns, ns)
        ns_map = merge_dicts(self.ns_map, ns_map)
        idx = first(idx, self.idx, 1)

        existing_elements = root.findall(qname(ns=ns_map.get(ns), name=name), ns_map)
        if idx > len(existing_elements) + 1:
            raise exc.SerializationError(
                "serialization can't be completed because {name}[{cur}] is going to be serialized, "
                "but {name}[{prev}] is not serialized.".format(name=name, cur=idx, prev=idx-1)
            )

        if obj is None:
            if self.required:
                raise exc.SerializationError("required element '{}' is not set".format(name))
            else:
                return None

        element = et.SubElement(root, qname(ns=ns_map.get(ns), name=name))
        element.text = encoder.encode(obj)

        return element

    def obj(self, xml, name=None, ns=None, ns_map=None, idx=None, full_path=()):
        name = first(self.name, name)
        ns = first(self.ns, ns)
        ns_map = merge_dicts(self.ns_map, ns_map)
        idx = first(idx, self.idx, 1)

        tag = tag_name(ns=ns, name=name, idx=idx)

        xml = xml.find(tag, ns_map)
        if xml is None or xml.text is None:
            if self.required:
                raise exc.DeserializationError("required element '/{}' not found".format('/'.join(full_path + (tag, ))))
            return None

        return xml.text


class WrapperXmlMapper(Mapper):
    """
    Wrapper mapper.
    """

    def __init__(self, path, wrapped, ns, ns_map, idx=None):
        self.ns = ns
        self.ns_map = ns_map
        self.name, *tail = path.split('/', 1)
        self.idx = idx
        self.required = wrapped.required

        if tail:
            self.wrapped = type(self)(tail[0], wrapped, ns, ns_map)
        else:
            self.wrapped = wrapped

    def xml(self, obj, root, name=None, ns=None, ns_map=None, idx=None, encoder=default_encoder):
        ns = first(self.ns, ns)
        ns_map = merge_dicts(self.ns_map, ns_map)
        idx = first(idx, self.idx, 1)

        existing_elements = root.findall(qname(ns=ns_map.get(ns), name=self.name), ns_map)
        if idx > len(existing_elements) + 1:
            raise exc.SerializationError(
                "serialization can't be completed because {name}[{cur}] is going to be serialized, "
                "but {name}[{prev}] is not serialized.".format(name=name, cur=idx, prev=idx - 1)
            )
        if idx == len(existing_elements) + 1:
            element = et.Element(qname(ns=ns_map.get(ns), name=self.name))
            new_element = True
        else:
            element = existing_elements[idx-1]
            new_element = False

        serialized = self.wrapped.xml(obj, element, name, ns, ns_map, encoder=encoder)
        if serialized is None:
            return None

        if new_element:
            root.append(element)

        return element

    def obj(self, xml, name=None, ns=None, ns_map=None, idx=None, full_path=()):
        ns = first(self.ns, ns)
        ns_map = merge_dicts(self.ns_map, ns_map)
        idx = first(idx, self.idx, 1)

        tag = tag_name(ns=ns, name=self.name, idx=idx)

        xml = xml.find(tag, ns_map)
        if xml is None:
            if self.wrapped.required:
                raise exc.DeserializationError("required element '/{}' not found".format('/'.join(full_path + (tag, ))))
            return None

        return self.wrapped.obj(xml, name, ns, ns_map, full_path=full_path + (tag,))


class ListXmlWrapper(Mapper):
    """
    Element list to XMl mapper. Implements methods for mapping a list of elements to a python list and vise versa.
    """

    def __init__(self, wrapped):
        self.wrapped = wrapped
        self.required = wrapped.required

    def xml(self, obj, root, name=None, ns=None, ns_map=None, _=None, encoder=default_encoder):
        name = first(self.wrapped.name, name)
        ns = first(self.wrapped.ns, ns)
        ns_map = merge_dicts(self.wrapped.ns_map, ns_map)

        children = []
        for idx, item in enumerate(obj or []):
            children.append(self.wrapped.xml(item, root, name, ns, ns_map, idx+1, encoder=encoder))

        return children or None

    def obj(self, xml, name=None, ns=None, ns_map=None, _=None, full_path=()):
        name = first(self.wrapped.name, name)
        ns = first(self.wrapped.ns, ns)
        ns_map = merge_dicts(self.wrapped.ns_map, ns_map)

        result = []
        for idx, e in enumerate(xml.iterfind(tag_name(ns=ns, name=name), ns_map)):
            result.append(self.wrapped.obj(xml, name, ns, ns_map, idx+1, full_path=full_path))

        return result


class ModelXmlMapper(Mapper):
    """
    Model to XMl mapper. Implements methods for mapping an xml element to a python object and vise versa.
    """

    def __init__(self, cls, name=None, ns=None, ns_map=None, idx=None, required=True):
        model_name, model_ns, model_ns_map, self.order = get_attrs(cls)
        self.cls = cls
        self.name = first(name, model_name, cls.__name__)
        self.ns = first(ns, model_ns)
        self.ns_map = merge_dicts(ns_map, model_ns_map)
        self.idx = idx
        self.required = required

    def xml(self, obj, root, name=None, ns=None, ns_map=None, idx=None, encoder=default_encoder):
        name = first(self.name, name)
        ns = first(self.ns, ns)
        ns_map = merge_dicts(self.ns_map, ns_map)
        idx = first(idx, self.idx, 1)

        existing_elements = root.findall(qname(ns=ns_map.get(ns), name=name), ns_map)
        if idx > len(existing_elements) + 1:
            raise exc.SerializationError(
                "serialization can't be completed because {name}[{cur}] is going to be serialized, "
                "but {name}[{prev}] is not serialized.".format(name=name, cur=idx, prev=idx-1)
            )

        if obj is None:
            if self.required:
                raise exc.SerializationError("required element '{}' is not set".format(name))
            else:
                return None

        element = et.Element(qname(ns=ns_map.get(ns), name=name))

        serialized_fields = []
        for field in reorder(attr.fields(self.cls), self.order, op.attrgetter('name')):
            mapper = field.metadata.get('paxb.mapper')
            if mapper:
                serialized = mapper.xml(getattr(obj, field.name), element, field.name, ns, ns_map, encoder=encoder)
                if serialized is not None:
                    serialized_fields.append(serialized)

        if not serialized_fields:
            return None
        else:
            root.append(element)
            return element

    def obj(self, xml, name=None, ns=None, ns_map=None, idx=None, full_path=()):
        name = first(self.name, name)
        ns = first(self.ns, ns)
        ns_map = merge_dicts(self.ns_map, ns_map)
        idx = first(idx, self.idx, 1)

        tag = tag_name(ns=ns, name=name, idx=idx)

        xml = xml.find(tag, ns_map)
        if xml is None:
            if self.required:
                raise exc.DeserializationError("required element '/{}' not found".format('/'.join(full_path + (tag, ))))
            return None

        cls_kwargs = {}

        for attr_field in attr.fields(self.cls):
            mapper = attr_field.metadata.get('paxb.mapper')
            if mapper:
                cls_kwargs[attr_field.name] = mapper.obj(xml, attr_field.name, ns, ns_map, full_path=full_path + (tag,))

        # Alter class initialization arguments that start with underscore (_). It is necessary because of
        # `attrs` library implementation specific. See https://www.attrs.org/en/stable/init.html#private-attributes.
        cls_kwargs = {k.lstrip('_'): w for k, w in cls_kwargs.items()}

        cls_kwargs = drop_nones(cls_kwargs)

        return self.cls(**cls_kwargs)



import base64
import typing as tp
import datetime as dt


def encode(value):
    """
    Default ``paxb`` value encoder. Encodes attributes or element text data during serialization.
    Supports :py:class:`str`, :py:class:`bytes`, :py:class:`datetime.date` and :py:class:`datetime.datetime` types.

    :param value: value to be encoded
    :return: encoded value string
    :rtype: str
    """

    if isinstance(value, tp.Text):
        return value
    if isinstance(value, tp.ByteString):
        return base64.b64encode(value)
    if isinstance(value, dt.datetime):
        return value.isoformat()
    if isinstance(value, dt.date):
        return value.isoformat()

    return str(value)



#!/usr/bin/env python

import sys
import setuptools.command.test
from setuptools import setup, find_packages

import paxb

requirements = [
    'attrs~=19.0'
]

test_requirements = [
    'pytest~=4.0',
    'xmldiff~=2.0'
]

with open('README.rst', 'r') as file:
    readme = file.read()


class PyTest(setuptools.command.test.test):
    user_options = [('pytest-args=', 'a', 'Arguments to pass to py.test')]

    def initialize_options(self):
        setuptools.command.test.test.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        import pytest
        sys.exit(pytest.main(self.pytest_args))


setup(
    name=paxb.__title__,
    version=paxb.__version__,
    description=paxb.__description__,
    long_description=readme,
    author=paxb.__author__,
    author_email=paxb.__email__,
    url=paxb.__url__,
    license=paxb.__license__,
    keywords=['xml', 'binding', 'mapping', 'serialization', 'deserialization'],
    python_requires=">=3.5",
    packages=find_packages(),
    install_requires=requirements,
    tests_require=test_requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: Public Domain',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    cmdclass={'test': PyTest},
)
